; 
;  Copyright (C) 2005, 2008 Greg Benison
; 
;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.
; 
;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.
; 
;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
; 

; derived from the BMRB restricted tabulation of protein chemical shifts.
(define-module (burrow chemical-shift-distributions)
  #:export (shift-lookup
	    shift-mean
	    shift-stddev
	    shift-probability
	    amino-acids
	    shift-top-probabilities))

; ---- 'shift' data type -----

; from the table of standard chemical shifts,
; return a 'shift object' of 'atom' of amino acid 'aa'
(define (shift-lookup aa atom)
  (let loop ((table stats-table))
    (cond ((null? table) #f)
	  ((and (eq? aa (caar table))
		(eq? atom (cadar table)))
	   (car table))
	  (else (loop (cdr table))))))

(define (shift-mean shift)
  (list-ref shift 6))

(define (shift-stddev shift)
  (list-ref shift 7))

; ---- manipulation functions -----
(define gauss-factor (sqrt (* 2 3.14159)))
(define (shift-probability-deprecated shift value)
  (let ((x (- (shift-mean shift) value))
	(dev (shift-stddev shift)))
    (/ (exp (- (expt (/ x (* 2 dev)) 2)))
       (* dev gauss-factor))))

; for an atom type defined by 'shift',
; what is the probability that its chemical shift
; would equal 'value'?
(define (shift-probability shift value)
  (let ((x (- (shift-mean shift) value))
	(dev (shift-stddev shift)))
    (exp (- (expt (/ x (* 2 dev)) 2)))))

; what is the probability that 'atom' of residue 'aa' will
; have chemical shift 'value'?
(define (shift-lookup-probability aa atom value)
  (let ((shift (shift-lookup aa atom)))
    (if shift (shift-probability shift value) 0.0)))

(define amino-acids
  '(ASP PRO LYS ILE TRP CYS GLY PHE GLN SER ASN VAL LEU TYR GLU ARG THR ALA MET HIS))

; Calculate the most probable residue type corresponding to a list of assigned chemical shifts.
; Return a list of (aa-type . probability), sorted in descending order
; of probability.
; 'args' is an alternating list of atom-type, chemical-shift-value
;
; e.g. shift-top-probabilities 'CA 55.2 'CB 32.5 'CG 22.4
(define (shift-top-probabilities . args)
  (define (prob aa args)
    (if (< (length args) 2)
	1.0
	(let ((atom-type (car args))
	      (value     (cadr args))
	      (rest      (cddr args)))
	  (* (shift-lookup-probability aa atom-type value)
	     (prob aa rest)))))
  (let* ((probabilities (map (lambda(aa)(prob aa args)) amino-acids))
	 (result        (map cons amino-acids probabilities)))
    (sort result (lambda(a b)(> (cdr a)(cdr b))))))

; ---- start BMRB table ----
(define stats-table
  '((     ALA     H        H        12913       3.53       11.58       8.18       0.60       )
    (     ALA     HA       H        10794       1.50        6.51       4.26       0.44       )
    (     ALA     HB       H        9980       -0.83        2.70       1.37       0.25       )
    (     ALA     C        C        6685      164.48      185.42     177.78       2.23       )
    (     ALA     CA       C        9643       43.00       65.52      53.19       2.02       )
    (     ALA     CB       C        8811        0.00       32.99      18.94       1.85       )
    (     ALA     N        N        10773      99.44      142.81     123.15       3.65       )
    (     ARG     H        H        8876        4.92       12.02       8.23       0.61       )
    (     ARG     HA       H        7410        1.34        6.44       4.29       0.45       )
    (     ARG     HB2      H        6490       -0.42        3.29       1.80       0.27       )
    (     ARG     HB3      H        5906       -0.20        3.29       1.78       0.28       )
    (     ARG     HG2      H        5710       -0.64        3.35       1.58       0.27       )
    (     ARG     HG3      H        5055       -0.40        3.35       1.56       0.28       )
    (     ARG     HD2      H        5500        0.83        4.69       3.12       0.24       )
    (     ARG     HD3      H        4784        1.22        4.28       3.12       0.24       )
    (     ARG     HE       H        2420        3.08       11.88       7.35       0.57       )
    (     ARG     HH11     H        293         5.82        9.45       6.83       0.39       )
    (     ARG     HH12     H        210         5.82        8.76       6.77       0.34       )
    (     ARG     HH21     H        271         4.59        9.70       6.79       0.40       )
    (     ARG     HH22     H        219         4.59        9.21       6.75       0.38       )
    (     ARG     C        C        4406      167.44      184.51     176.49       2.09       )
    (     ARG     CA       C        6464       45.00       67.98      56.86       2.39       )
    (     ARG     CB       C        5654       22.00       42.14      30.61       1.83       )
    (     ARG     CG       C        3323       18.22       37.83      27.26       1.31       )
    (     ARG     CD       C        3398       28.62       49.06      43.15       1.00       )
    (     ARG     CZ       C        96        156.49      160.60     159.18       0.95       )
    (     ARG     N        N        7064      103.60      137.60     120.75       3.83       )
    (     ARG     NE       N        1040        7.28      145.60      92.84      14.92       )
    (     ARG     NH1      N        44         70.10      112.50      73.35       6.17       )
    (     ARG     NH2      N        39         70.10      111.90      73.84       6.64       )
    (     ASP     H        H        10170       4.34       12.61       8.32       0.57       )
    (     ASP     HA       H        8473        2.59        6.34       4.60       0.32       )
    (     ASP     HB2      H        7639       -0.39        6.60       2.73       0.27       )
    (     ASP     HB3      H        7253        0.60        4.58       2.69       0.29       )
    (     ASP     C        C        5510      166.80      182.70     176.45       1.81       )
    (     ASP     CA       C        7822       39.80       62.49      54.64       2.09       )
    (     ASP     CB       C        7078       29.51       48.80      40.81       1.66       )
    (     ASP     CG       C        155       170.72      182.66     179.10       1.66       )
    (     ASP     N        N        8736      105.33      143.52     120.76       4.00       )
    (     ASN     H        H        7589        3.61       12.40       8.35       0.63       )
    (     ASN     HA       H        6323        2.61        6.39       4.68       0.37       )
    (     ASN     HB2      H        5769        0.10        4.47       2.81       0.33       )
    (     ASN     HB3      H        5527       -0.19        4.77       2.77       0.35       )
    (     ASN     HD21     H        4109        3.33        9.62       7.32       0.50       )
    (     ASN     HD22     H        4077        3.05       10.33       7.16       0.51       )
    (     ASN     C        C        3769      168.23      181.00     175.36       1.82       )
    (     ASN     CA       C        5484       41.31       62.13      53.50       1.96       )
    (     ASN     CB       C        4966       28.64       55.09      38.64       1.75       )
    (     ASN     CG       C        455       170.70      181.20     176.79       1.33       )
    (     ASN     N        N        6116      103.70      137.49     119.02       4.18       )
    (     ASN     ND2      N        2871      102.33      133.59     112.85       2.44       )
    (     CYS     H        H        4935        4.73       12.12       8.40       0.67       )
    (     CYS     HA       H        4539        1.64        6.43       4.69       0.56       )
    (     CYS     HB2      H        4312       -0.25        4.65       2.96       0.45       )
    (     CYS     HB3      H        4200       -0.83        4.65       2.92       0.49       )
    (     CYS     HG       H        20          0.25        7.39       2.36       2.08       )
    (     CYS     C        C        1497      168.50      182.73     174.83       2.05       )
    (     CYS     CA       C        2160       42.45       68.54      57.71       3.43       )
    (     CYS     CB       C        1888       20.10       62.07      34.01       6.63       )
    (     CYS     N        N        2629      100.48      135.89     119.99       4.63       )
    (     GLU     H        H        13211       4.31       12.17       8.33       0.61       )
    (     GLU     HA       H        11007       1.77        6.29       4.25       0.41       )
    (     GLU     HB2      H        9706        0.25        3.33       2.04       0.22       )
    (     GLU     HB3      H        8854        0.34        3.33       2.02       0.22       )
    (     GLU     HG2      H        8732        0.57        3.77       2.30       0.22       )
    (     GLU     HG3      H        7784        0.45        3.68       2.28       0.22       )
    (     GLU     C        C        7360      166.80      182.99     177.01       2.03       )
    (     GLU     CA       C        10368      46.80       66.60      57.45       2.13       )
    (     GLU     CB       C        9274       21.70       40.60      29.98       1.76       )
    (     GLU     CG       C        6018       25.10       51.92      36.04       1.34       )
    (     GLU     CD       C        153       173.41      185.20     181.20       3.47       )
    (     GLU     N        N        11475     104.54      138.60     120.69       3.66       )
    (     GLN     H        H        7186        5.41       11.94       8.21       0.59       )
    (     GLN     HA       H        6036        2.33        6.23       4.27       0.43       )
    (     GLN     HB2      H        5309       -0.14        4.00       2.05       0.26       )
    (     GLN     HB3      H        4861       -0.52        4.04       2.03       0.28       )
    (     GLN     HG2      H        4899        0.06        3.66       2.32       0.29       )
    (     GLN     HG3      H        4290        0.05        3.66       2.31       0.29       )
    (     GLN     HE21     H        3484        3.39       11.11       7.20       0.48       )
    (     GLN     HE22     H        3465        3.59        9.79       7.04       0.47       )
    (     GLN     C        C        3734      169.03      182.22     176.36       1.98       )
    (     GLN     CA       C        5467       46.82       65.58      56.58       2.17       )
    (     GLN     CB       C        4908       21.00       41.94      29.18       1.91       )
    (     GLN     CG       C        3204       21.64       41.70      33.74       1.18       )
    (     GLN     CD       C        399       171.70      183.50     179.71       1.24       )
    (     GLN     N        N        6073      105.26      139.55     119.86       3.77       )
    (     GLN     NE2      N        2659       92.49      133.30     111.89       1.98       )
    (     GLY     H        H        12739       3.01       12.22       8.34       0.69       )
    (     GLY     HA2      H        10511       0.94        7.31       3.96       0.38       )
    (     GLY     HA3      H        9893        1.27        6.01       3.91       0.38       )
    (     GLY     C        C        6291      165.80      183.00     174.04       1.90       )
    (     GLY     CA       C        9243       35.50       55.41      45.35       1.33       )
    (     GLY     N        N        10231      88.30      135.66     109.71       4.12       )
    (     HIS     H        H        3628        4.60       11.03       8.23       0.69       )
    (     HIS     HA       H        3049        1.95        8.90       4.62       0.47       )
    (     HIS     HB2      H        2732        0.37        8.70       3.10       0.39       )
    (     HIS     HB3      H        2613        0.52        8.70       3.07       0.40       )
    (     HIS     HD1      H        188         3.79       17.20       8.97       2.86       )
    (     HIS     HD2      H        2057        3.46        9.01       7.06       0.49       )
    (     HIS     HE1      H        1799        3.21       10.26       8.03       0.55       )
    (     HIS     HE2      H        96          6.62       16.53       9.99       2.77       )
    (     HIS     C        C        1895      166.90      180.80     175.35       2.08       )
    (     HIS     CA       C        2790       45.80       66.98      56.52       2.46       )
    (     HIS     CB       C        2532       23.40       43.30      30.23       2.16       )
    (     HIS     CG       C        49        122.67      136.80     130.81       3.25       )
    (     HIS     CD2      C        744        94.17      138.85     119.71       2.80       )
    (     HIS     CE1      C        546       115.52      144.00     137.00       2.32       )
    (     HIS     N        N        2985      105.00      136.48     119.47       4.23       )
    (     HIS     ND1      N        95        106.60      256.60     194.68      34.78       )
    (     HIS     NE2      N        107       114.71      231.50     175.85      18.81       )
    (     ILE     H        H        9020        3.75       11.49       8.28       0.70       )
    (     ILE     HA       H        7555        1.03        6.30       4.19       0.56       )
    (     ILE     HB       H        6900       -1.28        3.01       1.80       0.30       )
    (     ILE     HG12     H        6035       -2.02        2.69       1.28       0.39       )
    (     ILE     HG13     H        5690       -2.04        2.99       1.23       0.41       )
    (     ILE     HG2      H        6450       -2.07        1.77       0.80       0.27       )
    (     ILE     HD1      H        6384       -1.22        2.96       0.69       0.30       )
    (     ILE     C        C        4946      167.00      182.70     175.88       1.98       )
    (     ILE     CA       C        6971       51.15       71.70      61.59       2.77       )
    (     ILE     CB       C        6322       20.94       46.97      38.59       2.06       )
    (     ILE     CG1      C        4030        9.80       38.39      27.67       1.96       )
    (     ILE     CG2      C        4395        7.30       29.80      17.51       1.56       )
    (     ILE     CD1      C        4365        3.20       23.00      13.46       1.77       )
    (     ILE     N        N        7670       99.00      138.12     121.53       4.42       )
    (     LEU     H        H        14714       4.89       13.22       8.21       0.65       )
    (     LEU     HA       H        12036       1.58        6.24       4.31       0.47       )
    (     LEU     HB2      H        10587      -1.44        3.18       1.63       0.35       )
    (     LEU     HB3      H        9921       -1.21        3.18       1.55       0.37       )
    (     LEU     HG       H        9334       -1.06        3.90       1.52       0.34       )
    (     LEU     HD1      H        10282      -1.76        2.36       0.77       0.27       )
    (     LEU     HD2      H        9827       -1.16        2.78       0.74       0.29       )
    (     LEU     C        C        7923      167.49      189.78     177.00       2.06       )
    (     LEU     CA       C        11229      44.60       65.83      55.68       2.19       )
    (     LEU     CB       C        10098      29.70       51.29      42.24       1.90       )
    (     LEU     CG       C        6002       15.57       41.22      26.72       1.26       )
    (     LEU     CD1      C        6822       10.95       31.83      24.59       1.71       )
    (     LEU     CD2      C        6351       12.50       30.40      24.11       1.76       )
    (     LEU     N        N        12487     100.60      144.55     121.79       3.99       )
    (     LYS     H        H        13291       3.63       11.67       8.19       0.62       )
    (     LYS     HA       H        11057       0.68        6.04       4.27       0.44       )
    (     LYS     HB2      H        9691       -0.18        4.05       1.78       0.25       )
    (     LYS     HB3      H        8818       -0.58        3.95       1.76       0.27       )
    (     LYS     HG2      H        8451       -0.82        3.01       1.38       0.26       )
    (     LYS     HG3      H        7459       -0.98        2.99       1.36       0.28       )
    (     LYS     HD2      H        7314       -0.61        7.71       1.61       0.24       )
    (     LYS     HD3      H        6163       -0.10        3.61       1.60       0.24       )
    (     LYS     HE2      H        7121        1.09        4.32       2.92       0.20       )
    (     LYS     HE3      H        5900        1.03        4.32       2.92       0.21       )
    (     LYS     HZ       H        645         2.82        9.90       7.49       0.43       )
    (     LYS     C        C        6682      121.16      185.00     176.68       2.15       )
    (     LYS     CA       C        9662       46.10       63.14      56.96       2.25       )
    (     LYS     CB       C        8670       21.00       46.60      32.73       1.84       )
    (     LYS     CG       C        5307       17.30       40.50      24.90       1.24       )
    (     LYS     CD       C        4974       21.20       42.60      28.87       1.24       )
    (     LYS     CE       C        4748       25.98       56.00      41.81       0.97       )
    (     LYS     N        N        10883     101.10      140.30     121.04       3.92       )
    (     LYS     NZ       N        9          32.09      131.04      65.83      46.43       )
    (     MET     H        H        3557        4.87       11.89       8.25       0.60       )
    (     MET     HA       H        3038        1.86        6.35       4.39       0.48       )
    (     MET     HB2      H        2635       -0.99        4.07       2.04       0.33       )
    (     MET     HB3      H        2443       -0.99        3.15       2.01       0.35       )
    (     MET     HG2      H        2360       -0.36        4.40       2.44       0.35       )
    (     MET     HG3      H        2186       -0.47        4.24       2.41       0.39       )
    (     MET     HE       H        1531       -0.21        3.30       1.88       0.39       )
    (     MET     C        C        1903      168.20      181.47     176.31       2.08       )
    (     MET     CA       C        2882       47.70       63.76      56.20       2.25       )
    (     MET     CB       C        2562       24.86       46.46      33.02       2.24       )
    (     MET     CG       C        1536       20.83       37.96      32.03       1.33       )
    (     MET     CE       C        1071        0.00       42.57      17.27       2.65       )
    (     MET     N        N        3068      106.30      134.80     120.05       3.64       )
    (     PHE     H        H        6547        5.21       12.03       8.38       0.72       )
    (     PHE     HA       H        5404        2.13        6.56       4.61       0.58       )
    (     PHE     HB2      H        4948        0.17        4.32       2.99       0.38       )
    (     PHE     HB3      H        4744        0.30        4.31       2.96       0.39       )
    (     PHE     HD1      H        4200        2.46        8.08       7.06       0.33       )
    (     PHE     HD2      H        3249        2.46        8.15       7.06       0.33       )
    (     PHE     HE1      H        3781        4.10        8.80       7.10       0.33       )
    (     PHE     HE2      H        2996        4.10        8.80       7.10       0.32       )
    (     PHE     HZ       H        2815        4.53        9.50       7.03       0.43       )
    (     PHE     C        C        3499      167.70      181.96     175.56       2.02       )
    (     PHE     CA       C        4956       47.31       69.82      58.23       2.68       )
    (     PHE     CB       C        4495       25.52       50.40      39.93       2.10       )
    (     PHE     CG       C        34        127.84      140.40     136.82       3.31       )
    (     PHE     CD1      C        1521      118.50      135.90     131.44       1.36       )
    (     PHE     CD2      C        903       119.10      137.20     131.44       1.28       )
    (     PHE     CE1      C        1258      114.75      133.90     130.57       1.58       )
    (     PHE     CE2      C        742       121.30      133.90     130.67       1.09       )
    (     PHE     CZ       C        935       117.51      138.60     129.22       1.80       )
    (     PHE     N        N        5513      102.76      139.02     120.66       4.25       )
    (     PRO     H2       H        1           8.51        8.51       8.51       0.00       )
    (     PRO     HA       H        5771        1.63        6.05       4.39       0.34       )
    (     PRO     HB2      H        5265       -0.78        4.02       2.06       0.36       )
    (     PRO     HB3      H        5009       -0.36        3.79       2.02       0.37       )
    (     PRO     HG2      H        4614       -0.93        4.92       1.93       0.34       )
    (     PRO     HG3      H        4166       -0.90        4.92       1.91       0.36       )
    (     PRO     HD2      H        4832        0.63        5.36       3.64       0.36       )
    (     PRO     HD3      H        4555       -0.26        5.36       3.62       0.39       )
    (     PRO     C        C        3509      170.50      182.30     176.75       1.56       )
    (     PRO     CA       C        5201       50.12       72.28      63.31       1.62       )
    (     PRO     CB       C        4620       25.48       50.25      31.78       1.26       )
    (     PRO     CG       C        2855       19.31       33.79      27.20       1.16       )
    (     PRO     CD       C        2913       26.96       66.22      50.32       1.15       )
    (     PRO     N        N        149       104.39      145.26     131.17       9.61       )
    (     SER     H        H        10722       2.96       13.13       8.29       0.60       )
    (     SER     HA       H        9073        1.43        6.46       4.49       0.42       )
    (     SER     HB2      H        7995        1.53        5.41       3.88       0.27       )
    (     SER     HB3      H        7223        1.49        5.01       3.86       0.29       )
    (     SER     HG       H        147         0.00        8.97       5.33       1.08       )
    (     SER     C        C        5402      165.00      182.20     174.66       1.78       )
    (     SER     CA       C        7912       45.13       68.40      58.69       2.19       )
    (     SER     CB       C        6960       55.50       73.90      63.81       1.56       )
    (     SER     N        N        8638      102.60      133.10     116.25       3.76       )
    (     THR     H        H        9874        5.54       11.39       8.25       0.63       )
    (     THR     HA       H        8215        0.87        6.36       4.47       0.48       )
    (     THR     HB       H        7324        0.92        8.35       4.17       0.34       )
    (     THR     HG1      H        293         0.32        8.21       4.74       1.76       )
    (     THR     HG2      H        7222       -0.83        4.07       1.15       0.24       )
    (     THR     C        C        4898      165.50      182.02     174.60       1.79       )
    (     THR     CA       C        7195       51.61       72.80      62.18       2.73       )
    (     THR     CB       C        6411       30.16       81.53      69.63       1.86       )
    (     THR     CG2      C        4268       11.70       38.90      21.48       1.26       )
    (     THR     N        N        8150       97.70      134.00     115.52       4.93       )
    (     TRP     H        H        2111        4.49       11.09       8.29       0.80       )
    (     TRP     HA       H        1748        2.28        6.75       4.70       0.54       )
    (     TRP     HB2      H        1601        0.42        4.54       3.21       0.36       )
    (     TRP     HB3      H        1533        0.77        4.49       3.15       0.36       )
    (     TRP     HD1      H        1479        4.90        8.93       7.15       0.35       )
    (     TRP     HE1      H        1491        5.16       12.14      10.10       0.56       )
    (     TRP     HE3      H        1339        4.89        8.98       7.35       0.41       )
    (     TRP     HZ2      H        1404        4.90        8.60       7.30       0.33       )
    (     TRP     HZ3      H        1302        3.88        8.20       6.87       0.44       )
    (     TRP     HH2      H        1315        4.92       10.90       7.00       0.42       )
    (     TRP     C        C        1005      169.63      181.83     176.14       1.99       )
    (     TRP     CA       C        1449       44.69       69.76      57.67       2.64       )
    (     TRP     CB       C        1321       22.48       43.02      30.03       2.03       )
    (     TRP     CG       C        86        102.53      114.60     110.21       1.84       )
    (     TRP     CD1      C        542       109.30      132.35     126.39       2.11       )
    (     TRP     CD2      C        73        120.20      130.70     127.46       1.36       )
    (     TRP     CE2      C        57        118.37      177.71     137.67       6.64       )
    (     TRP     CE3      C        453       113.97      137.60     120.60       2.25       )
    (     TRP     CZ2      C        510        81.81      123.76     114.13       2.02       )
    (     TRP     CZ3      C        455       114.90      138.39     121.37       1.84       )
    (     TRP     CH2      C        468        91.62      130.07     123.60       2.21       )
    (     TRP     N        N        1609      101.75      134.89     121.65       4.50       )
    (     TRP     NE1      N        905       106.75      139.96     129.45       2.30       )
    (     TYR     H        H        5799        4.39       11.92       8.33       0.74       )
    (     TYR     HA       H        4865        1.20        6.73       4.62       0.57       )
    (     TYR     HB2      H        4470        0.31        4.70       2.91       0.37       )
    (     TYR     HB3      H        4286        0.34        4.70       2.88       0.38       )
    (     TYR     HD1      H        4145        4.98        8.53       6.95       0.30       )
    (     TYR     HD2      H        3293        4.43       10.50       6.95       0.31       )
    (     TYR     HE1      H        4023        4.58        7.86       6.71       0.23       )
    (     TYR     HE2      H        3213        4.56        8.50       6.71       0.24       )
    (     TYR     HH       H        94          5.31       13.75       9.13       1.56       )
    (     TYR     C        C        2663      168.37      182.40     175.45       2.03       )
    (     TYR     CA       C        4004       45.20       65.80      58.10       2.61       )
    (     TYR     CB       C        3567       28.95       57.73      39.27       2.18       )
    (     TYR     CG       C        69        120.94      138.70     129.02       2.82       )
    (     TYR     CD1      C        1439      116.44      136.70     132.66       1.34       )
    (     TYR     CD2      C        794       113.00      136.70     132.47       1.74       )
    (     TYR     CE1      C        1412      110.70      133.31     117.84       1.43       )
    (     TYR     CE2      C        785       114.58      133.31     117.86       1.66       )
    (     TYR     CZ       C        64        153.54      162.70     156.47       2.25       )
    (     TYR     N        N        4484      103.60      144.96     120.72       4.46       )
    (       VAL     H        H        11522       3.98       11.63       8.28       0.70       )
    (       VAL     HA       H        9584        0.97        6.24       4.17       0.58       )
    (       VAL     HB       H        8706       -0.56        3.42       1.99       0.32       )
    (       VAL     HG1      H        8494       -1.09        2.56       0.84       0.26       )
    (       VAL     HG2      H        8178       -2.32        3.32       0.82       0.29       )
    (       VAL     C        C        6084      168.50      181.72     175.69       1.94       )
    (       VAL     CA       C        8752       51.10       70.02      62.49       2.95       )
    (       VAL     CB       C        7847       22.04       42.84      32.66       1.82       )
    (       VAL     CG1      C        5467       11.60       29.00      21.45       1.46       )
    (       VAL     CG2      C        5167       11.30       31.10      21.32       1.63       )
    (       VAL     N        N        9833       96.29      143.29     121.06       4.78       )))
